
""" Contains functions to cluster spheres objects in memory; used in subpocket clustering. """

from .spheres import Spheres
import itertools
import logging
import numpy as np
import scipy

logger = logging.getLogger(__name__)

def cluster_within_r(spheres, radius, allow_new=True):
    """ Cluster spheres with the same radius using DBSCAN, modifying input data in situ

    Args:
      spheres (Spheres): complete set of input spheres
      radius (float): radius at which clustering is to occur
      allow_new (bool): permit new clusters? (Default value = True)

    """
    from sklearn.cluster import DBSCAN

    r_indices = np.where(spheres.r == radius)[0]
    selected = spheres.xyzrg[r_indices, :]

    ungrouped_indices = np.where(selected[:, 4] == 0)[0]

    if ungrouped_indices.shape[0] > 0:
        ungrouped_data = selected[ungrouped_indices]
        db = DBSCAN(eps=radius, min_samples=1).fit(ungrouped_data[:, 0:3])

        if allow_new:
            starting_index = np.amax(selected[:, 4]) + 1
            np.put(spheres.g, r_indices[ungrouped_indices], db.labels_ + starting_index)
        else:
            starting_index = np.amin(selected[:, 4]) - 1
            np.put(spheres.g, r_indices[ungrouped_indices], -1 * db.labels_ + starting_index)

    logger.debug("Clustered spheres at radius {0}".format(radius))


def cluster_between_r(spheres, ref_radius, target_radius):
    """ Cluster spheres from a target radius to a reference radius, modifying input data in situ

    Args:
      spheres (Spheres): complete set of input spheres
      ref_radius (float): radius from which cluster identities will be drawn
      target_radius (float): radius to which cluster identities will be propagated

    """
    ref_data = spheres.xyzrg[spheres.r == ref_radius]
    r_indices = np.where(spheres.r == target_radius)[0]
    target_data = spheres.xyzrg[r_indices]

    if (ref_data.shape[0] > 0) and (target_data.shape[0] > 0):

        kdtree = scipy.spatial.cKDTree(ref_data[:, 0:3])
        dist, indices = kdtree.query(target_data[:, 0:3], distance_upper_bound=ref_radius, n_jobs=-1)

        target_indices = indices < ref_data.shape[0]
        ref_indices = indices[target_indices]

        np.put(spheres.g, r_indices[target_indices], ref_data[ref_indices, 4])

    logger.debug("Clustered spheres at radius {0} against those at {1}".format(target_radius, ref_radius))


def cluster_improperly_grouped(spheres, radius, min_cluster_size=1, max_clusters=None):
    """ Reassigns improperly clustered spheres to 'proper' clusters, modifying input data in situ

    Args:
      spheres (Spheres): complete set of input spheres
      radius (float): radius at which closest groups are identified
      min_cluster_size (int): minimum number of spheres in a 'proper' cluster (Default value = 1)
      max_clusters (int): maximum number of 'proper' clusters (Default value = None)

    """
    min_group = np.amin(spheres.g)
    group_list = np.flip(np.arange(min_group, 0), axis=0)
    reassign_groups_to_closest(spheres, group_list, radius)

    spheres.remove_ungrouped()

    group_counts = np.bincount(spheres.g.astype(int))
    small_groups = np.where(group_counts < min_cluster_size)[0]
    if len(small_groups) > 1:
        # always includes the 0 group
        reassign_groups_to_closest(spheres, small_groups[1:], radius)
    disconnected_small_groups = np.where(group_counts < min_cluster_size)[0]
    if len(disconnected_small_groups) > 1:
        spheres.remove_groups(disconnected_small_groups)

    group_counts = np.bincount(spheres.g.astype(int))
    num_groups = np.count_nonzero(group_counts)

    if max_clusters is not None:
        if num_groups > max_clusters:
            reassign_groups_to_closest(spheres, np.where(group_counts > 0)[0], radius, iterations=(num_groups - max_clusters))
    logger.debug("Improperly grouped spheres re-clustered yielding {0} groups".format(num_groups))


def extract_groups(spheres, surf_radius=None, prefix=None, group_names=None):
    """ Extracts spheres belonging to each cluster from the complete input set and optionally calculates bounded surfaces

    Args:
      spheres (Spheres): complete set of input spheres
      surf_radius: radius used to calculate bounding spheres for individual groups (Default value = None)
      prefix: prefix to identify new surfaces (Default value = None)

    Returns:
      group_list ([Spheres]): a list of Spheres objects each corresponding to a different cluster

    """
    groups = np.unique(spheres.g)

    group_list = []
    for index, group in enumerate(groups):
        group_spheres = Spheres(xyzrg = spheres.xyzrg[spheres.g == group].copy())
        if prefix is not None:
            group_spheres.name = "{0}_p0_sp{1}".format(prefix, index)
        elif group_names is not None:
            group_spheres.name = group_names[index]
        group_list.append(group_spheres)

    logger.debug("Extracting {0} groups from {1}".format(len(group_list), spheres.name))

    if surf_radius is not None:
        exterior_list = [group_spheres.calculate_surface(probe_radius=surf_radius)[0] for group_spheres in group_list]
        reindices = np.flip(np.argsort([s.mesh.volume for s in exterior_list]), 0)

        new_group_list = []
        new_ext_list = []
        for index in reindices:
            g_s = group_list[index]
            e_s = exterior_list[index]

            g_s.g = index
            e_s.g = index
            g_s.mesh = e_s.mesh.copy()

            e_s.name = g_s.name
            new_group_list.append(g_s)
            new_ext_list.append(e_s)

        return new_group_list
    else:
        return group_list


def hierarchically_cluster_spheres(spheres, ordered_radii, min_new_radius=None, min_cluster_size=10, max_clusters=None):
    """ Cluster spheres by grouping spheres at large radius and propagating those assignments down to smaller radii

    Args:
      spheres (Spheres): complete set of input spheres
      ordered_radii ([float]): list of radii ordered from largest to smallest
      min_new_radius (float): smallest spheres to keep (Default value = None)
      min_cluster_size (int): minimum number of spheres in a cluster (Default value = 10)
      max_clusters (int): maximum number of clusters (Default value = None)

    """
    if min_new_radius is None:
        min_new_radius = np.amin(ordered_radii)

    for index, radius in enumerate(ordered_radii):
        initial_grouped = spheres.xyzrg[spheres.g != 0].shape[0]
        if index > 0:
            cluster_between_r(spheres, ref_radius=ordered_radii[index - 1], target_radius=ordered_radii[index])

        cluster_within_r(spheres, radius, allow_new=(radius >= min_new_radius))
    logger.debug("Finished naive sphere clustering for spheres in {0}".format(spheres.name))

    cluster_improperly_grouped(spheres, radius=ordered_radii[-1], min_cluster_size=min_cluster_size, max_clusters=max_clusters)
    logger.debug("Finished hierarchically clustering for spheres in {0}".format(spheres.name))


def identify_closest_grouped(spheres, group, radius):
    """ Identifies the closest 'properly' grouped cluster to a specified group

    Args:
      spheres (Spheres): complete set of input spheres
      group (float): group for which to identify the closest clusters
      radius (float): radius at which to perform the search

    Returns:
      group (float): passthrough of input group
      closest (float): id of the closest cluster
      magnitude (int): number of pairwise closest connections between the queried group and the closest identified cluster
    """
    target_indices = np.where((spheres.r == radius) & (spheres.g == group))[0]
    grouped_indices = np.where((spheres.r == radius) & (spheres.g > 0) & (spheres.g != group))[0]

    target_data = spheres.xyzrg[target_indices]
    grouped_data = spheres.xyzrg[grouped_indices]

    if (target_data.shape[0] > 0) and (grouped_data.shape[0] > 0):
        kdtree = scipy.spatial.cKDTree(grouped_data[:, 0:3])
        dist, indices = kdtree.query(target_data[:, 0:3], distance_upper_bound=1.41 * radius, n_jobs=-1)
        # 1.41 factor allows the two spheres to intersect at pi/4 from the closest point

        t_indices = indices < grouped_data.shape[0]
        group_indices = indices[t_indices]
        if len(group_indices) > 0:
            counts = np.bincount(grouped_data[group_indices,4].astype(int))
            closest = np.argmax(counts)
            magnitude = counts[closest]
            return [group, closest, magnitude]
        else:
            return [None, None, 0]
    else:
        return [None, None, 0]


def merge_sphere_list(s_list, r=None, g=None):
    """

    Args:
      s_list ([Spheres]): list of input spheres
      r (float): radius value to assign to output Spheres (Default value = None)
      g (float): group value to assign to output Spheres (Default value = None)

    Returns:
      merged_spheres (Spheres): a single Spheres object containing the merged input lists
    """
    selected_data_list = []

    for i, s in enumerate(s_list):
        if s is None:
            continue
        selected_data = s.xyzrg

        if r is not None:
            selected_data = selected_data[selected_data[:, 3] == r]
        if g is not None:
            selected_data = selected_data[selected_data[:, 4] == g]

        if selected_data.shape[0] > 0:
            selected_data_list.append(selected_data)

    if len(selected_data_list) > 0:
        return Spheres(xyzrg=np.vstack(selected_data_list))
    else:
        return None


def reassign_group(spheres, source_group, target_group):
    """ Reassign a group in place

    Args:
      spheres (Spheres): complete set of input spheres
      source_group (float): group to change
      target_group (float): new group id

    """
    source_indices = np.where(spheres.g == source_group)

    np.put(spheres.g, source_indices, target_group)


def reassign_groups_to_closest(spheres, group_list, radius, iterations=None, preserve_largest=False):
    """ Reassign a group to the closest group as identified by maximum linkage; operates in place

    Args:
      spheres (Spheres): complete set of input spheres
      group_list ([float]): list of group ids which are to be iteratively reassigned
      radius (float): radius at which searches are to take place
      iterations (int): number of times to attempt to reassign groups (Default value = None)
      preserve_largest: keep the group id of the group with more members? (Default value = False)

    """
    if iterations is None:
        iterations = len(group_list)

    for i in range(iterations):
        linkages = []
        for group in group_list:
            linkages.append(identify_closest_grouped(spheres, group, radius))

        nonzero_linkages = [link for link in linkages if link[2] > 0]
        if len(nonzero_linkages) > 0:
            best_link = sorted(nonzero_linkages, key=lambda x: x[2])[-1]
            if preserve_largest:
                group_sizes = np.bincount(spheres.g.astype(int))
                if group_sizes[best_link[0]] > group_sizes[best_link[1]]:
                    best_link = [best_link[1], best_link[0]]

            reassign_group(spheres, best_link[0], best_link[1])
        else:
            break


def remove_interior(spheres):
    """ Remove all spheres which are completely enclosed in larger spheres; operates in place

    Args:
      spheres (Spheres): complete set of input spheres

    """
    min_rad = np.amin(spheres.r)
    max_rad = np.amax(spheres.r)

    point_tree = scipy.spatial.cKDTree(spheres.xyz)
    neighbors = point_tree.query_ball_tree(point_tree, r=(max_rad - min_rad))

    interior_indices = []
    for point_index, nlist in enumerate(neighbors):
        if point_index in interior_indices:
            continue

        if len(nlist) <= 1:
            continue

        inclusion = spheres.r[point_index] - spheres.r[nlist].reshape(-1, 1) - scipy.spatial.distance.cdist(spheres.xyz[nlist], spheres.xyz[point_index].reshape(1, -1))
        included_indices = np.where(inclusion > 0)[0]
        if len(included_indices) > 0:
            interior_indices.extend(list(np.array(nlist)[included_indices]))

    interior_indices = np.unique(interior_indices).astype(int)
    spheres.xyzrg = np.delete(spheres.xyzrg, interior_indices, axis=0)

    logger.debug("Removed interior spheres from {0}".format(spheres.name))


def remove_included_spheres(spheres, ref_spheres, radius):
    """ Removes all spheres with centers within radius of ref_spheres

    """

    kdtree = scipy.spatial.cKDTree(spheres.xyz)
    groups = kdtree.query_ball_point(ref_spheres.xyz, radius, n_jobs=-1)
    indices = np.unique(list(itertools.chain.from_iterable(groups)))

    spheres.xyzrg = np.delete(spheres.xyzrg, indices, axis=0)

    logger.debug("Removed all spheres within {0} A of reference".format(radius))



def remove_overlap(spheres, radii=None, spacing=0.1, iterations=20, tolerance=0.02, static_last_group=False):
    """ Remove overlap between groups; operates in place

    Args:
      spheres (Spheres): complete set of input spheres
      radii ([float]): radii at which to perform searches for overlap (Default value = None)
      spacing (float): binning radius (Default value = 0.1)
      iterations (int): number of times to attempt overlap removal (Default value = 20)
      tolerance (float): overlap tolerance (Default value = 0.02)
      static_last_group (bool): don't move the 'other' group but rather the first group twice as much (effectively leaves the group with the highest index in place while moving everything else around it)

    """
    from sklearn.preprocessing import normalize
    groups = np.unique(spheres.g)[:-1]

    if spheres.xyzrg.shape[0] == 0:
        logger.warning("Attempting to remove overlap in an empty sphere set")
        return

    if radii is None:
        radii = [np.amax(spheres.r)]
        spacing = radii[0]

    for radius in radii:
        for group in groups:
            group_indices = np.where((spheres.g == group) & (spheres.r > (radius - spacing)) & (spheres.r <= radius))[0]
            other_indices = np.where((spheres.g != group) & (spheres.r > (radius - spacing)) & (spheres.r <= radius))[0]

            if len(group_indices) == 0 or len(other_indices) == 0:
                continue

            group_data = spheres.xyzrg[group_indices]
            other_data = spheres.xyzrg[other_indices]

            other_tree = scipy.spatial.cKDTree(other_data[:, 0:3])
            group_tree = scipy.spatial.cKDTree(group_data[:, 0:3])

            neighbors = group_tree.query_ball_tree(other_tree, r=2 * radius)

            altered_group_indices = []
            altered_other_indices = []

            for iteration in range(iterations):
                overlaps = np.zeros(len(neighbors))
                overlap_indices = -1 * np.ones(len(neighbors))

                for group_index, nlist in enumerate(neighbors):
                    if len(nlist) == 0:
                        continue
                    overlap = other_data[nlist, 3].reshape(-1, 1) + group_data[group_index, 3] - scipy.spatial.distance.cdist(other_data[nlist, 0:3], group_data[group_index, 0:3].reshape(1, -1))
                    most_overlapping_index = np.argmax(overlap)
                    if overlap[most_overlapping_index] > 0:
                        overlaps[group_index] = overlap[most_overlapping_index]
                        overlap_indices[group_index] = nlist[most_overlapping_index]

                overlapped_group_indices = np.where(overlaps > tolerance)[0]
                if len(overlapped_group_indices) == 0:
                    break

                overlaps = overlaps[overlapped_group_indices]
                overlap_indices = overlap_indices[overlapped_group_indices].astype(int)

                reorder = np.argsort(overlaps)[::-1]
                overlaps = overlaps[reorder]
                overlap_indices = overlap_indices[reorder]
                overlapped_group_indices = overlapped_group_indices[reorder]

                foo, closest_indices = np.unique(overlap_indices, return_index=True)
                overlaps = overlaps[closest_indices]
                overlap_indices = overlap_indices[closest_indices]
                overlapped_group_indices = overlapped_group_indices[closest_indices]

                if not static_last_group:
                    overlap_adjustment = 0.26 * overlaps # 0.25 should work but leads to a logarithmic approach of proper adjustment
                else:
                    overlap_adjustment = 0.51 * overlaps # move the mobile group twice as much if the other group isn't moving

                vector = overlap_adjustment[:, np.newaxis] * normalize(group_data[overlapped_group_indices, 0:3] - other_data[overlap_indices, 0:3])

                group_data[overlapped_group_indices, 0:3] = group_data[overlapped_group_indices, 0:3] + vector
                group_data[overlapped_group_indices, 3] = group_data[overlapped_group_indices, 3] - overlap_adjustment
                altered_group_indices.extend(list(overlapped_group_indices))

                if not static_last_group:
                    other_data[overlap_indices, 0:3] = other_data[overlap_indices, 0:3] - vector
                    other_data[overlap_indices, 3] = other_data[overlap_indices, 3] - overlap_adjustment
                    altered_other_indices.extend(list(overlap_indices))

            altered_group_indices = np.unique(altered_group_indices).astype(int)
            altered_other_indices = np.unique(altered_other_indices).astype(int)

            spheres.xyzrg[group_indices[altered_group_indices]] = group_data[altered_group_indices]
            spheres.xyzrg[other_indices[altered_other_indices]] = other_data[altered_other_indices]




""" Handles options for PyVOL. Reads configuration files and objects and converts them to option dictionaries and then back again. Accepts and parses string input for all parameters from commandline/gui inputs. """

from . import utilities
import configparser
import logging
import numpy as np
import os
import re
import tempfile
from datetime import datetime

logger = logging.getLogger(__name__)

def clean_opts(input_opts):
    """ Cleans opts and then returns the sanitized contents

    Args:
      input_opts (dict): dictionary of all input options

    Returns:
      cleaned_opts (dict): dictionary containing all options for a PyVOL run with extraneous options removed and necessary defaults provided
    """

    timestamp = datetime.now().strftime(r"%H%M%S-%f")

    trimmed_opts = {}
    for k, v in input_opts.items():
        if v is not None:
            trimmed_opts[k] = v
    input_opts = trimmed_opts

    # Load options
    opts = {}
    opts["protein"] = input_opts.get("protein")
    opts["prot_file"] = input_opts.get("prot_file")
    opts["lig_file"] = input_opts.get("lig_file")
    opts["min_rad"] = float(input_opts.get("min_rad", 1.4))
    opts["max_rad"] = float(input_opts.get("max_rad", 3.4))
    opts["constrain_radii"] = input_opts.get("constrain_inputs", False)

    opts["mode"] = input_opts.get("mode")
    opts["coordinates"] = input_opts.get("coordinates")
    if opts.get("coordinates") is not None:
        if isinstance(opts.get("coordinates"), str):
            try:
                opts["coordinates"] = np.asarray([float(x) for x in opts.get("coordinates").split(",")]).reshape(-1,3)
            except:
                logger.error("Coordinates argument not parsed from str correctly: {0}".format(opts.get("coordinates")))
                raise ValueError
        if isinstance(opts.get("coordinates"), list):
            opts["coordinates"] = np.array([float(x) for x in opts.get("coordinates")]).reshape(-1,3)

        if opts.get("coordinates").shape != (1,3):
            logger.error("Coordinates argument contains the wrong number of dimensions: {0}".format(opts.get("coordinates").shape))
            raise ValueError
    opts["resid"] = input_opts.get("resid")
    opts["lig_excl_rad"] = input_opts.get("lig_excl_rad")
    if opts.get("lig_excl_rad") is not None:
        opts["lig_excl_rad"] = float(opts["lig_excl_rad"])
    opts["lig_incl_rad"] = input_opts.get("lig_incl_rad")
    if opts.get("lig_incl_rad") is not None:
        opts["lig_incl_rad"] = float(opts["lig_incl_rad"])
    opts["min_volume"] = input_opts.get("min_volume")
    if opts.get("min_volume") is not None:
        opts["min_volume"] = float(input_opts.get("min_volume"))
    else:
        if opts.get("mode") == "all":
            opts["min_volume"] = 200

    opts["subdivide"] = input_opts.get("subdivide", False)
    if not isinstance(opts["subdivide"], bool):
        opts["subdivide"] = False
    if opts["subdivide"]:
        opts["max_clusters"] = int(input_opts.get("max_clusters", 50))
        opts["min_subpocket_rad"] = float(input_opts.get("min_subpocket_rad", 1.7))
        opts["max_subpocket_rad"] = float(input_opts.get("max_subpocket_rad", 3.4))
        opts["min_subpocket_surf_rad"] = float(input_opts.get("min_subpocket_surf_rad", 1.0))
        opts["radial_sampling"] = float(input_opts.get("radial_sampling", 0.1))
        opts["inclusion_radius_buffer"] = float(input_opts.get("inclusion_radius_buffer", 1.0))
        opts["min_cluster_size"] = int(input_opts.get("min_cluster_size", 50))

    opts["project_dir"] = input_opts.get("project_dir")
    opts["output_dir"] = input_opts.get("output_dir")
    opts["prefix"] = input_opts.get("prefix")
    if opts["prefix"] is None:
        if opts.get("prot_file") is not None:
            opts["prefix"] = "{0}_{1}".format(timestamp, os.path.splitext(os.path.basename(opts["prot_file"]))[0])
        elif opts.get("protein") is not None:
            opts["prefix"] = "{0}_{1}".format(timestamp, opts.get("protein").split()[0].strip("(").strip(")"))
        else:
            logger.error("No protein input detected: either prot_file or the PyMOL protein selection must be defined")
            raise ValueError("No protein geometry defined: provide either a protein file or protein PyMOL selection")

    if opts.get("output_dir") is None:
        if opts.get("project_dir") is not None:
            opts["output_dir"] = os.path.join(opts.get("project_dir"), "{0}.pyvol".format(opts.get("prefix")))
        else:
            opts["output_dir"] = os.path.join(os.getcwd(), "{0}.pyvol".format(opts.get("prefix")))

    utilities.check_dir(opts.get("output_dir"))

    if opts.get("prot_file") is None:
        opts["prot_file"] = os.path.join(opts.get("output_dir"), "{0}_prot.pdb".format(opts.get("prefix")))

    if (opts.get("ligand") is not None) and (opts.get("lig_file") is None):
        opts["lig_file"] = os.path.join(opts.get("output_dir"), "{0}_lig.pdb".format(opts.get("prefix")))


    opts["logger_stream_level"] = input_opts.get("logger_stream_level")
    if opts["logger_stream_level"] not in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
        opts["logger_stream_level"] = None
    opts["logger_file_level"] = input_opts.get("logger_file_level")
    if opts["logger_file_level"] not in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
        opts["logger_file_level"] = None

    opts["protein"] = input_opts.get("protein")
    opts["ligand"] = input_opts.get("ligand")
    opts["protein_only"] = input_opts.get("protein_only")
    if not isinstance(opts["protein_only"], bool):
        opts["protein_only"] = False
    opts["display_mode"] = input_opts.get("display_mode")
    if not opts["display_mode"] in ["solid", "mesh", "spheres"]:
        opts["display_mode"] = "solid"
    opts["palette"] = input_opts.get("palette")
    if opts.get("palette") is not None:
        palette_valid = False
        if isinstance(opts.get("palette"), str):
            fragments = re.split(r"[\(\)]", opts.get("palette"))
            cleaned_pieces = []
            for fragment in fragments:
                pieces = list(filter(None, fragment.split(",")))

                if len(pieces) > 0:
                    try:
                        rgb = [float(piece) for piece in pieces]
                        cleaned_pieces.append(rgb)
                    except:
                        cleaned_pieces.extend(pieces)
            opts["palette"] = cleaned_pieces

    if opts.get("alpha") is not None:
        opts["alpha"] = float(input_opts.get("alpha"))
    else:
        opts["alpha"] = 1.0

    # Clean options
    if opts["prot_file"] is None:
        logger.error("A protein file must be provided--Terminating job")
        raise

    if opts["constrain_radii"]:
        if opts["min_rad"] < 1.2:
            logger.info("Minimum radius constrained from {0} to 1.2".format(opts["min_rad"]))
            opts["min_rad"] = 1.2
        elif opts["min_rad"] > 2.0:
            logger.info("Minimum radius constrained from {0} to 2.0".format(opts["min_rad"]))
            opts["min_rad"] = 2.0

        if opts["max_rad"] < 2.0:
            logger.info("Maximum radius constrained from {0} to 2.0".format(opts["max_rad"]))
            opts["max_rad"] = 2.0
        elif opts["max_rad"] > 5.0:
            logger.info("Maximum radius constrained from {0} to 5.0".format(opts["max_rad"]))
            opts["max_rad"] = 5.0

    if opts["mode"] in ["all", "largest"]:
        opts["lig_file"] = None
        opts["coordinates"] = None
        opts["resid"] = None
        opts["lig_excl_rad"] = None
        opts["lig_incl_rad"] = None
    else:
        if opts["lig_file"] is not None:
            if opts["mode"] is None:
                logger.info("Pocket identified through provided ligand")
            opts["mode"] = "specific"
            opts["resid"] = None
            opts["coordinates"] = None
        elif opts["resid"] is not None:
            if opts["mode"] is None:
                logger.info("Pocket identified through resid: {0}".format(opts["resid"]))
            opts["mode"] = "specific"
            opts["coordinates"] = None
        elif opts["coordinates"] is not None:
            logger.info("Pocket identified through coordinates: {0}".format(opts["coordinates"]))
            opts["mode"] = "specific"
        else:
            opts["mode"] = "largest"

    if opts["subdivide"]:
        if opts["max_clusters"] <= 1:
            logger.warning("Subpocket analysis impossible with maximum clusters of {0}; disabling subpocket analysis".format(opts["max_clusters"]))
            opts["subdivide"] = False
            opts["max_clusters"] = None
            opts["min_subpocket_rad"] = None
            opts["max_cluster_rad"] = None
            opts["min_subpocket_surf_rad"] = None
            opts["radial_sampling"] = None
            opts["inclusion_radius_buffer"] = None
            opts["min_cluster_size"] = None

    # Remove all empty options
    cleaned_opts = {}
    for k, v in opts.items():
        if v is not None:
            cleaned_opts[k] = v

    logger.debug("Input options sanitized")
    return cleaned_opts


def opts_to_cfg(opts):
    """ creates the configuration file corresponding to input options

    Args:
      opts (dict): option dictionary for which to create a configuration object

    Returns:
      config (ConfigParser): configuration object containing formatted options
    """

    config = configparser.ConfigParser()

    config.add_section("General")
    if opts.get("prot_file") is not None:
        config.set("General", "prot_file", str(opts.get("prot_file")))
    if opts.get("lig_file") is not None:
        config.set("General", "lig_file", str(opts.get("lig_file")))
    if opts.get("min_rad") is not None:
        config.set("General", "min_rad", str(opts.get("min_rad")))
    if opts.get("max_rad") is not None:
        config.set("General", "max_rad", str(opts.get("max_rad")))
    if opts.get("constrain_radii") is not None:
        config.set("General", "constrain_radii", str(opts.get("constrain_radii")))

    config.add_section("Specification")
    if opts.get("mode") is not None:
        config.set("Specification", "mode", str(opts.get("mode")))
    if opts.get("resid") is not None:
        config.set("Specification", "resid", str(opts.get("resid")))
    if opts.get("coordinates") is not None:
        config.set("Specification", "coordinates", str(opts.get("coordinates")))
    if opts.get("lig_excl_rad") is not None:
        config.set("Specification", "lig_excl_rad", str(opts.get("lig_excl_rad")))
    if opts.get("lig_incl_rad") is not None:
        config.set("Specification", "lig_incl_rad", str(opts.get("lig_incl_rad")))
    if opts.get("min_volume") is not None:
        config.set("Specification", "min_volume", str(opts.get("min_volume")))

    config.add_section("Partitioning")
    if opts.get("subdivide") is not None:
        config.set("Partitioning", "subdivide", str(opts.get("subdivide")))
    if opts.get("max_clusters") is not None:
        config.set("Partitioning", "max_clusters", str(opts.get("max_clusters")))
    if opts.get("min_subpocket_rad") is not None:
        config.set("Partitioning", "min_subpocket_rad", str(opts.get("min_subpocket_rad")))
    if opts.get("max_subpocket_rad") is not None:
        config.set("Partitioning", "max_subpocket_rad", str(opts.get("max_subpocket_rad")))
    if opts.get("min_subpocket_surf_rad") is not None:
        config.set("Partitioning", "min_subpocket_surf_rad", str(opts.get("min_subpocket_surf_rad")))
    if opts.get("radial_sampling") is not None:
        config.set("Partitioning", "radial_sampling", str(opts.get("radial_sampling")))
    if opts.get("inclusion_radius_buffer") is not None:
        config.set("Partitioning", "inclusion_radius_buffer", str(opts.get("inclusion_radius_buffer")))
    if opts.get("min_cluster_size") is not None:
        config.set("Partitioning", "min_cluster_size", str(opts.get("min_cluster_size")))

    config.add_section("Output")
    if opts.get("project_dir") is not None:
        config.set("Output", "project_dir", str(opts.get("project_dir")))
    if opts.get("output_dir") is not None:
        config.set("Output", "output_dir", str(opts.get("output_dir")))
    if opts.get("prefix") is not None:
        config.set("Output", "prefix", str(opts.get("prefix")))
    if opts.get("logger_stream_level") is not None:
        config.set("Output", "logger_stream_level", str(opts.get("logger_stream_level")))
    if opts.get("logger_file_level") is not None:
        config.set("Output", "logger_file_level", str(opts.get("logger_file_level")))


    config.add_section("PyMOL")
    if opts.get("protein") is not None:
        config.set("PyMOL", "protein", str(opts.get("protein")))
    if opts.get("ligand") is not None:
        config.set("PyMOL", "ligand", str(opts.get("ligand")))
    if opts.get("protein_only") is not None:
        config.set("PyMOL", "protein_only", str(opts.get("protein_only")))
    if opts.get("display_mode") is not None:
        config.set("PyMOL", "display_mode", str(opts.get("display_mode")))
    if opts.get("palette") is not None:
        config.set("PyMOL", "palette", str(opts.get("palette")))
    if opts.get("alpha") is not None:
        config.set("PyMOL", "alpha", str(opts.get("alpha")))

    return config

def defaults_to_cfg():
    """ Creates a blank template cfg with all accepted fields and reasonable default values

    Returns:
      config (ConfigParser): configuration object containing defaults
    """
    config = configparser.ConfigParser(allow_no_value=True)
    config.add_section("General")
    config.set("General", "prot_file")
    config.set("General", "lig_file")
    config.set("General", "min_rad", "1.4")
    config.set("General", "max_rad", "3.4")
    config.set("General", "constrain_radii", "True")

    config.add_section("Specification")
    config.set("Specification", "mode", "largest")
    config.set("Specification", "coordinates")
    config.set("Specification", "resid")
    config.set("Specification", "lig_excl_rad")
    config.set("Specification", "lig_incl_rad")
    config.set("Specification", "min_volume", "200")

    config.add_section("Partitioning")
    config.set("Partitioning", "subdivide", "False")
    config.set("Partitioning", "max_clusters")
    config.set("Partitioning", "min_subpocket_rad", "1.7")
    config.set("Partitioning", "max_subpocket_rad", "3.4")
    config.set("Partitioning", "min_subpocket_surf_rad", "1.0")
    config.set("Partitioning", "radial_sampling", "0.1")
    config.set("Partitioning", "inclusion_radius_buffer", "1.0")
    config.set("Partitioning", "min_cluster_size", "50")

    config.add_section("Output")
    config.set("Output", "project_dir")
    config.set("Output", "prefix")
    config.set("Output", "logger_stream_level", "INFO")
    config.set("Output", "logger_file_level", "DEBUG")

    config.add_section("PyMOL")
    config.set("PyMOL", "protein")
    config.set("PyMOL", "ligand")
    config.set("PyMOL", "protein_only", "False")
    config.set("PyMOL", "display_mode", "solid")
    config.set("PyMOL", "palette")
    config.set("PyMOL", "alpha", "0.85")

    return config


def defaults_to_file(filename):
    """ writes a default configuation file to disk

    Args:
      filename (str): output filename to which to write the configuration file to disk
    """

    cfg_to_file(defaults_to_cfg(), filename)


def cfg_to_opts(config):
    """ converts a config to opts

    Args:
      config (ConfigParser): configuration object from which options are to be extracted

    Returns:
      opts (dict): dictionary of options read in from the configuration object
    """

    opts = {}
    opts["prot_file"] = config.get("General", "prot_file", fallback=None)
    opts["lig_file"] = config.get("General", "lig_file", fallback=None)
    opts["min_rad"] = config.getfloat("General", "min_rad", fallback=1.4)
    opts["max_rad"] = config.getfloat("General", "max_rad", fallback=3.4)

    opts["mode"] = config.get("Specification", "mode", fallback=None)
    opts["resid"] = config.get("Specification", "resid", fallback=None)
    opts["coordinates"] = config.get("Specification", "coordinates", fallback=None)
    opts["lig_excl_rad"] = config.getfloat("Specification", "lig_excl_rad", fallback=-1)
    opts["lig_incl_rad"] = config.getfloat("Specification", "lig_incl_rad", fallback=-1)

    opts["subdivide"] = config.getboolean("Partitioning", "subdivide", fallback=False)
    opts["min_volume"] = config.getint("Partitioning", "min_volume", fallback=200)
    opts["max_clusters"] = config.getint("Partitioning", "max_clusters", fallback=100)
    opts["min_subpocket_rad"] = config.getfloat("Partitioning", "min_subpocket_rad", fallback=1.7)
    opts["max_subpocket_rad"] = config.getfloat("Partitioning", "max_subpocket_rad", fallback=3.4)
    opts["min_subpocket_surf_rad"] = config.getfloat("Partitioning", "min_subpocket_surf_rad", fallback=1.0)
    opts["radial_sampling"] = config.getfloat("Partitioning", "radial_sampling", fallback=0.1)
    opts["inclusion_radius_buffer"] = config.getfloat("Partitioning", "inclusion_radius_buffer", fallback=1.0)
    opts["min_cluster_size"] = config.getint("Partitioning", "min_cluster_size", fallback=50)

    opts["project_dir"] = config.get("Output", "project_dir", fallback=None)
    opts["output_dir"] = config.get("Output", "output_dir", fallback=None)
    opts["prefix"] = config.get("Output", "prefix", fallback=None)
    opts["logger_stream_level"] = config.get("Output", "logger_stream_level", fallback="INFO")
    opts["logger_file_level"] = config.get("Output", "logger_file_level", fallback="DEBUG")

    opts["protein"] = config.get("PyMOL", "protein", fallback=None)
    opts["ligand"] = config.get("PyMOL", "ligand", fallback=None)
    opts["protein_only"] = config.get("PyMOL", "protein_only", fallback=True)
    opts["display_mode"] = config.get("PyMOL", "display_mode", fallback="solid")
    opts["palette"] = config.get("PyMOL", "palette", fallback=None)
    opts["alpha"] = config.get("PyMOL", "alpha", fallback=0.85)

    return opts


def cfg_to_file(cfg, filename):
    """ writes a configuration to file

    Args:
      cfg (ConfigParser): configuration object to be written to disk
      filename (str): target filename on disk
    """

    with open(filename, 'w') as configfile:
        cfg.write(configfile)
        logger.info("Configuration file written to {0}".format(filename))


def file_to_cfg(filename):
    """ reads a cfg file into a configuration object

    Args:
      filename (str): input filename of a configuration file

    Returns:
      config (ConfigParser): configuration object holding the contents of the file
    """

    config = configparser.ConfigParser()
    config.read(filename)
    logger.info("Configuration file read from {0}".format(filename))
    return config


def file_to_opts(filename):
    """ reads a cfg file and converts it into an options dictionary

    Args:
      filename (str): input filename of a configuration file

    Returns:
      opts (dict): dictionary object containing PyVOL options
    """

    return cfg_to_opts(file_to_cfg(filename))


def opts_to_file(opts, filename=None):
    """ writes options to a configuration file

    Args:
      opts (dict): dictionary object containing PyVOL options
      filename (str): target file to which to write the configuration

    """

    if filename is None:
        filename = os.path.join(opts.get("output_dir"), "{0}.cfg".format(opts.get("prefix")))

    cfg_to_file(opts_to_cfg(opts), filename)




# Experimental code not included in distributed release

from pyvol.spheres import Spheres
import logging
import numpy as np
import pandas as pd
import scipy

logger = logging.getLogger(__name__)

try:
    from rdkit import Chem
    from rdkit.Chem.Draw import rdMolDraw2D
    from rdkit.Chem import AllChem
except:
    logger.warning("rdkit not found; small molecule analysis will not be available prior to installation")


# def pose_report(pose_file, pocket_file, output_dir, output_prefix=None, name_parameter="_Name", scoring_parameter="r_i_glide_gscore", pocket_tolerance=3, panelx=250, panely=200, molsPerRow=4, rowsPerPage=6, palette=[(1,0.2,0.2), (1,0.55,0.15), (1,1,0.2), (0.2,1,0.2), (0.3,0.3,1), (0.5,1,1), (1,0.5,1)]):
#     """ Creates a report that highlights 2D compound representations by subpocket occupancy according to the poses in a provided sdf file
#
#     Args:
#       pose_file (str): input SDF file containing docked compound poses
#       pocket_file (str): input csv containing the spheres 5 dimensional array describing subpocket geometry; output with a "_spa.csv" ending
#       output_dir (str): output directory for all files
#       output_prefix (str): output prefix
#       name_parameter (str): SDF property key for the molecule name
#       scoring_parameter (str): SDF property key for whichever property should be shown in the report
#       pocket_tolerance (float): maximum distance (Angstrom) at which an atom outside of the defined subpocket volumes is still associated with a subpocket
#       panelx (int): horizontal width of the drawing space for each molecule
#       panely (int): vertical height of the drawing space for each molecule
#       molsPerRow (int): number of molecules to fit on a row (total width is <= panelx * molsPerRow)
#       rowsPerPage (int): number of rows of molecules to fit on each page (total height is <= panely * rowsPerPage)
#       palette ([(float,float,float)]): list of tuples of fractional RGB values that controls the highlighting colors
#
#     """
#
#     poses.compound_occupancy(pose_file, pocket_file, output_dir, output_prefix, name_parameter, scoring_parameter, pocket_tolerance, panelx, panely, molsPerRow, rowsPerPage, palette)


def compound_occupancy(pose_file, pocket_file, output_dir, output_prefix=None, name_parameter="_Name", scoring_parameter="r_i_glide_gscore", pocket_tolerance=3, panelx=250, panely=200, molsPerRow=4, rowsPerPage=6, palette=[(1,0.2,0.2), (1,0.55,0.15), (1,1,0.2), (0.2,1,0.2), (0.3,0.3,1), (0.5,1,1), (1,0.5,1)]):
    """ Creates a report that highlights 2D compound representations by subpocket occupancy according to the poses in a provided sdf file

    Args:
      pose_file (str): input SDF file containing docked compound poses
      pocket_file (str): input csv containing the spheres' 5 dimensional array describing subpocket geometry; output with a "_spa.csv" ending
      output_dir (str): output directory for all files
      output_prefix (str): output prefix
      name_parameter (str): SDF property key for the molecule name
      scoring_parameter (str): SDF property key for whichever property should be shown in the report
      pocket_tolerance (float): maximum distance (Angstrom) at which an atom outside of the defined subpocket volumes is still associated with a subpocket
      panelx (int): horizontal width of the drawing space for each molecule
      panely (int): vertical height of the drawing space for each molecule
      molsPerRow (int): number of molecules to fit on a row (total width is <= panelx * molsPerRow)
      rowsPerPage (int): number of rows of molecules to fit on each page (total height is <= panely * rowsPerPage)
      palette ([(float,float,float)]): list of tuples of fractional RGB values that controls the highlighting colors

    """

    molecules = [x for x in Chem.ForwardSDMolSupplier(pose_file) if x is not None]

    spheres = np.loadtxt(spheres_file, delimiter=' ')
    subpocket_spheres = Spheres(xyzrg=sp)

    for molecule in molecules:
        set_atom_subpockets(molecule, subpocket_spheres.propagate_groups_to_external(molecule.GetConformers()[0].GetPositions(), tolerance=pocket_tolerance))
    draw_molecules_with_subpockets(molecules, output_dir, output_prefix, palette, panelx, panely, molsPerRow, rowsPerPage, name_parameter, scoring_parameter)


def draw_molecules_with_subpockets(molecules, output_dir, output_prefix, palette=[(1,0.2,0.2), (1,0.55,0.15), (1,1,0.2), (0.2,1,0.2), (0.3,0.3,1), (0.5,1,1), (1,0.5,1)], panelx=250, panely=200, molsPerRow=4, rowsPerPage=6, name_parameter="_Name", scoring_parameter="r_i_glide_gscore"):
    """ Creates a report that highlights the 2D representation of docked molecules to indicate subpocket occupancy

    Args:
      molecules ([rdkit.Chem.rdchem.Mol]): list of rdkit molecules to draw
      output_dir (str): output directory
      output_prefix (str): output prefix
      palette ([(float,float,float)]): list of tuples of fractional RGB values; TODO: allow input of pymol style color strings
      panelx (int): horizontal width of the drawing space for each molecule
      panely (int): vertical height of the drawing space for each molecule
      molsPerRow (int): number of molecules to fit on a row (total width is <= panelx * molsPerRow)
      rowsPerPage (int): number of rows of molecules to fit on each page (total height is <= panely * rowsPerPage)
      name_parameter (str): SDF property key for the molecule name
      scoring_parameter (str): SDF property key for whichever property should be shown in the report

    """


    hats = []
    hbnds = []
    hatcolors = []
    hbndcolors = []
    prepped_mols = []
    legends = []
    for index, m in enumerate(molecules):
        ats = []
        bnds = []
        atcolors = {}
        bndcolors = {}

        for atom in m.GetAtoms():
            subpocket = atom.GetIntProp("subpocket")
            idx = atom.GetIdx()
            if subpocket >= 0:
                ats.append(idx)
            atcolors[idx] = palette[subpocket]

        for bnd in m.GetBonds():
            if bnd.GetBeginAtom().GetIntProp("subpocket") == bnd.GetEndAtom().GetIntProp("subpocket"):
                bnds.append(bnd.GetIdx())
                bndcolors[bnd.GetIdx()] = palette[bnd.GetBeginAtom().GetIntProp("subpocket")]

        hats.append(ats)
        hbnds.append(bnds)
        hatcolors.append(atcolors)
        hbndcolors.append(bndcolors)
        AllChem.Compute2DCoords(m)
        prepped_mols.append(rdMolDraw2D.PrepareMolForDrawing(m))
        legends.append("{0}: {1} ({2})".format(index, m.GetProp(name_parameter), m.GetProp(scoring_parameter)))

    molsPerPage = molsPerRow * rowsPerPage
    nPages = len(molecules) // molsPerPage
    if molsPerPage * nPages < len(molecules):
        nPages += 1

    page_data = []
    for page in range(nPages):
        start_mol = page * molsPerPage
        end_mol = min((page + 1) * molsPerPage, len(molecules))

        nRows = (end_mol - start_mol) // molsPerRow
        if molsPerRow * nRows < (end_mol - start_mol):
            nRows += 1

        canvasx = panelx * molsPerRow
        canvasy = panely * nRows

        drawer = rdMolDraw2D.MolDraw2DCairo(canvasx, canvasy, panelx, panely)

        page_mols = prepped_mols[start_mol:end_mol]
        page_hats = hats[start_mol:end_mol]
        page_hbnds = hbnds[start_mol:end_mol]
        page_hatcolors = hatcolors[start_mol:end_mol]
        page_hbndcolors = hbndcolors[start_mol:end_mol]
        page_legends = legends[start_mol:end_mol]

        drawer.DrawMolecules(page_mols, highlightAtoms=page_hats,highlightBonds=page_hbnds, highlightAtomColors=page_hatcolors, highlightBondColors=page_hbndcolors, legends=page_legends)

        drawer.FinishDrawing()
        txt = drawer.GetDrawingText()
        page_data.append(txt)

    filenames = [os.path.join(output_dir, "{0}_table_{1}.png".format(output_prefix, x)) for x in range(len(page_data))]

    for index, pd in enumerate(page_data):
        with open(filenames[index], 'wb') as f:
            f.write(page_data[index])


def set_atom_subpockets(molecule, groups):
    """ Apply calculated subpocket groups to each atom in a molecule

    Args:
      molecule (rdkit.Chem.rdchem.Mol): rdkit molecule object
      groups ([int]): list of subpocket identifiers

    """
    if molecule.GetNumAtoms() != groups.shape[0]:
        raise ValueError("Incorrect number of groups for molecule")

    for index, atom in enumerate(molecule.GetAtoms()):
        atom.SetIntProp("subpocket", int(groups[index]))




""" PyMOL convenience functions used by the front-end contained in pymol_interface. """

import logging

logger = logging.getLogger(__name__)

try:
    from pymol import cgo, cmd, CmdException
except:
    logger.warning("PyMOL not imported")

def construct_palette(color_list=None, max_value=7, min_value=1):
    """ Construct a palette

    Args:
      color_list ([str]): list of PyMOL color strings (Default value = None)
      max_value (int): max palette index (Default value = 7)
      min_value (int): min palette index (Default value = 1)

    Returns:
      palette ([str]): list of color definitions

    """
    if color_list is None:
        color_list = ['tv_red', 'tv_orange', 'tv_yellow', 'tv_green', 'tv_blue', 'aquamarine', 'violet']
    if max_value <= len(color_list):
        logger.debug("Input palette accepted")
        return color_list[:max_value]

    colors = [cmd.get_color_tuple(x) for x in color_list]
    output_range = max_value - min_value + 1

    palette = []

    color_vectors = len(colors) - 1
    steps = output_range / color_vectors
    for cv in range(color_vectors):
        for x in range(steps):
            fx = float(x) / steps
            cl = [fx * colors[cv + 1][i] + (1 - fx) * colors[cv][i] for i in range(3)]
            cn = '0x%02x%02x%02x' % tuple([255 * x for x in cl])
            palette.append(cn)
    logger.debug("Palette constructed with {0} colors".format(len(palette)))
    return palette


def display_pseudoatom_group(spheres, name, color='gray60', palette=None):
    """ Displays a collection of pseudoatoms

    Args:
      spheres (Spheres): Spheres object holding pocket geometry
      name (str): display name
      color (str): PyMOL color string (Default value = 'gray60')
      palette ([str]): palette (Default value = None)

    """

    if spheres is None:
        return

    for index, xyzrg in enumerate(spheres.xyzrg):
        if palette is None:
            cmd.pseudoatom("{0}.{1}".format(name, index), pos=list(xyzrg[0:3]), vdw=float(xyzrg[3]), color=color)
        else:
            cmd.pseudoatom("{0}.{1}".format(name, index), pos=list(xyzrg[0:3]), vdw=float(xyzrg[3]), color=palette[int(xyzrg[4] - 1)])

    group_name = "{0}_g".format(name)
    cmd.group(group_name, "{0}.*".format(name))
    cmd.show("spheres", group_name)
    logger.debug("Pseudoatom group of {0} spheres created with group name {1}".format(spheres.xyzrg.shape[0], group_name))
    return group_name


def display_spheres_object(spheres, name, state=1, color='marine', alpha=1.0, mode="solid", palette=None):
    """ Loads a mesh object into a cgo list for display in PyMOL

    Args:
      spheres (Spheres): Spheres object containing all geometry
      name (str): display name
      state (int): model state (Default value = 1)
      color (str): PyMOL color string (Default value = 'marine')
      alpha (float): transparency value (Default value = 1.0)
      mode (str): display mode (Default value = "solid")
      palette ([str]): palette (Default value = None)

    """

    alpha = float(alpha)
    print(alpha)
    if spheres is None:
        return None

    if (mode == "mesh") or (mode == "solid"):
        if spheres.mesh is None:
            return None
        else:
            if mode == "solid":
                cmd.load_cgo(mesh_to_solid_CGO(spheres.mesh, color=color, alpha=alpha), name, state)
            else:
                cmd.load_cgo(mesh_to_wireframe_CGO(spheres.mesh, color=color, alpha=alpha), name, state)
            return None
    elif mode == "spheres":
        return display_pseudoatom_group(spheres, name, color=color, palette=None)


def mesh_to_solid_CGO(mesh, color='gray60', alpha=1.0):
    """Creates a solid CGO object for a mesh for display in PyMOL

    Args:
      mesh (Trimesh): Trimesh mesh object
      color (str): PyMOL color string (Default value = 'gray60')
      alpha (float): transparency value (Default value = 1.0)

    Returns:
      cgobuffer (str): CGO buffer that contains the instruction to load a solid object

    """
    print(alpha)
    cgobuffer = [cgo.BEGIN, cgo.TRIANGLES, cgo.ALPHA, alpha]
    color_values = cmd.get_color_tuple(cmd.get_color_index(color))

    for face in mesh.faces:
        for v_index in face:
            cgobuffer.append(cgo.COLOR)
            cgobuffer.extend(color_values)

            cgobuffer.append(cgo.NORMAL)
            cgobuffer.extend([mesh.vertex_normals[v_index][i] for i in range(3)])
            cgobuffer.append(cgo.VERTEX)
            cgobuffer.extend([mesh.vertices[v_index][i] for i in range(3)])
    cgobuffer.append(cgo.END)
    logger.debug("CGO solid object created for mesh: {0}".format(mesh))
    return cgobuffer


def mesh_to_wireframe_CGO(mesh, color='gray60', alpha=1.0):
    """Creates a wireframe CGO object for a mesh for display in PyMOL

    Args:
      mesh (Trimesh): Trimesh mesh object
      color (str): PyMOL color string (Default value = 'gray60')
      alpha (float): transparency value (Default value = 1.0)

    Returns:
      cgobuffer (str): CGO buffer that contains the instruction to load a wireframe object

    """

    cgobuffer = [cgo.BEGIN, cgo.LINES, cgo.ALPHA, alpha]

    cgobuffer.append(cgo.COLOR)
    cgobuffer.extend(cmd.get_color_tuple(cmd.get_color_index(color)))

    for edge in mesh.edges:
        cgobuffer.append(cgo.VERTEX)
        cgobuffer.extend(mesh.vertices[edge[0]])
        cgobuffer.append(cgo.VERTEX)
        cgobuffer.extend(mesh.vertices[edge[1]])

    cgobuffer.append(cgo.END)
    logger.debug("CGO wireframe object created for mesh: {0}".format(mesh))
    return cgobuffer




from .spheres import Spheres
from . import cluster, configuration, utilities
import glob
import inspect
import itertools
import logging
import numpy as np
import os
import pandas as pd
import shutil
import sys

logger = logging.getLogger(__name__)

def load_calculation(data_dir, input_opts=None):
    """ load the results of a calculation from file

    Args:
      data_dir (str): directory where previous calculation results are stored
      input_opts (dict): dictionary of pyvol options that is used to update the options read in from file

    Returns:
      pockets ([Spheres]): a list of Spheres objects each of which contains the geometric information describing a distinct pocket or subpocket
      opts (dict): updated PyVOL options dictionary

    """

    if not os.path.isdir(data_dir):
        logger.error("{0} is not a directory".format(data_dir))
        raise FileNotFoundError

    cfg_files = glob.glob(os.path.join(data_dir, "*.cfg"))
    if len(cfg_files) == 0:
        logger.error("No cfg file found in {0}".format(data_dir))
        raise FileNotFoundError
    elif len(cfg_files) > 1:
        logger.error("Multiple cfg files found in {0}".format(data_dir))
        raise FileNotFoundError

    opts = configuration.file_to_opts(cfg_files[0])
    if isinstance(input_opts, dict):
        opts.update(input_opts)
        opts = configuration.clean_opts(opts)

    rept_file = os.path.join(data_dir, "{0}.rept".format(opts.get("prefix")))
    if not os.path.isfile(rept_file):
        logger.error("No rept file found at {0}".format(rept_file))
        raise FileNotFoundError

    rept_df = pd.read_csv(rept_file)
    pockets = []
    for index, row in rept_df.iterrows():
        xyzrg_file = os.path.join(data_dir, "{0}.xyzrg".format(row["name"]))
        pockets.append(Spheres(spheres_file=xyzrg_file))

    return pockets, opts


def pocket(**opts):
    """Calculates the SES for a binding pocket

    Args:
      opts (dict): dictionary containing all PyVOL options (see pyvol.pymol_interface.pymol_pocket_cmdline for details)

    Returns:
      pockets ([Spheres]): a list of Spheres objects each of which contains the geometric information describing a distinct pocket or subpocket

    """

    if os.path.dirname(opts.get("prot_file")) != opts.get("output_dir"):
        new_prot_file = os.path.join(opts.get("output_dir"), os.path.basename(opts.get("prot_file")))
        shutil.copyfile(opts.get("prot_file"), new_prot_file)
        opts["prot_file"] = new_prot_file

        if opts.get("lig_file") is not None:
            new_lig_file = os.path.join(opts.get("output_dir"), os.path.basename(opts.get("lig_file")))
            shutil.copyfile(opts.get("lig_file"), new_lig_file)
            opts["lig_file"] = new_lig_file

    p_s = Spheres(pdb=opts.get("prot_file"), name="{0}_prot".format(opts.get("prefix")))
    logger.debug("Protein geometry read from {0}".format(opts.get("prot_file")))

    pl_s = p_s.copy()
    if opts.get("lig_file") is not None:
        l_s = Spheres(pdb=opts.get("lig_file"), r=opts.get("lig_incl_rad"), name="{0}_lig_incl".format(opts.get("prefix")))
        logger.debug("Ligand geometry read from {0}".format(opts.get("lig_file")))
        if opts.get("lig_incl_rad") is not None:
            pl_s = p_s + l_s
            logger.debug("Ligand-inclusion radius of {0} applied".format(opts.get("lig_incl_rad")))
    else:
        l_s = None

    pl_s.name = "{0}_interior".format(opts.get("prefix"))

    pl_bs = pl_s.calculate_surface(probe_radius=opts.get("max_rad"))[0]
    logger.debug("Outer bulk-solvent surface calculated")
    pl_bs.name = "{0}_boundary".format(opts.get("prefix"))

    pa_s = p_s + pl_bs
    pa_s.name = "{0}_exterior".format(opts.get("prefix"))
    if (l_s is not None) and (opts.get("lig_excl_rad") is not None):
        le_s = Spheres(xyz=l_s.xyzr, r=opts.get("lig_excl_rad"), name="{0}_lig_excl".format(opts.get("prefix")))
        le_bs = le_s.calculate_surface(probe_radius=opts.get("max_rad"))[0]
        pa_s = pa_s + le_bs
        logger.debug("Ligand-excluded radius of {0} applied".format(opts.get("lig_excl_rad")))

    if opts.get("mode") == "all":
        all_pockets = pa_s.calculate_surface(probe_radius=opts.get("min_rad"), all_components=True, min_volume=opts.get("min_volume"))
        for index, pocket in enumerate(all_pockets):
            pocket.name = "{0}_p{1}".format(opts.get("prefix"), index)
        logger.info("Pockets calculated using mode 'all': {0}".format(len(all_pockets)))
        if opts.get("subdivide"):
            logger.warning("Subpocket clustering not currently supported when calculating all independent pockets")
    else:
        if opts.get("mode") == "largest":
            bp_bs = pa_s.calculate_surface(probe_radius=opts.get("min_rad"), all_components=True, largest_only=True)[0]
            logger.info("Largest pocket identified")
        elif opts.get("mode") == "specific":
            if opts.get("coordinates") is not None:
                coordinate = opts.get("coordinates")
                logger.info("Specific pocket identified from coordinate: {0}".format(opts.get("coordinates")))
            elif opts.get("resid") is not None:
                resid = str(opts.get("resid"))
                chain = None
                if not resid[0].isdigit():
                    chain = resid[0]
                    resid = int(resid[1:])
                else:
                    resid = int(resid)
                coordinate = utilities.coordinates_for_resid(opts.get("prot_file"), resid=resid, chain=chain)
                logger.info("Specific pocket identified from residue: {0} -> {1} (truncated)".format(opts.get("resid"), coordinate[0,:]))
            elif l_s is not None:
                lig_coords = l_s.xyz
                coordinate = np.mean(l_s.xyz, axis=0).reshape(1, -1)
                logger.info("Specific pocket identified from mean ligand position: {0}".format(coordinate))
            else:
                logger.error("A coordinate, ligand, or residue must be supplied to run in specific mode")
                return None

            p_bs = p_s.calculate_surface(probe_radius=opts.get("min_rad"))[0]
            id_coord = p_bs.nearest_coord_to_external(coordinate).reshape(1, -1)
            bp_bs = pa_s.calculate_surface(probe_radius=opts.get("min_rad"), coordinate=id_coord)[0]
        else:
            logger.error("Unrecognized mode <{0}>--should be 'all', 'largest', or 'specific'".format(opts.get("mode")))
            return None

        bp_bs.name = "{0}_p0".format(opts.get("prefix"))

        if bp_bs.mesh.volume > pl_bs.mesh.volume:
            logger.error("Binding pocket not correctly identified--try an alternative method to specify the binding pocket")
            return [], opts
        else:
            all_pockets = [bp_bs]

        if opts.get("subdivide"):
            all_pockets.extend(subpockets(bounding_spheres = pa_s, ref_spheres = bp_bs, **opts))
            logger.info("Subpockets identified: {0}".format(len(all_pockets) - 1))

    write_report(all_pockets, **opts)
    write_cfg(**opts)

    return all_pockets, opts


def pocket_wrapper(**opts):
    """ wrapper for pocket that configures the logger, sanitizes inputs, and catches errors; useful when running from the command line or PyMOL but split from the core code for programmatic usage

    Args:
      opts (dict): dictionary containing all PyVOL options (see pyvol.pymol_interface.pymol_pocket_cmdline for details)

    Returns:
      pockets ([Spheres]): a list of Spheres objects each of which contains the geometric information describing a distinct pocket or subpocket
      output_opts (dict): dictionary containing the actual options used in the pocket calculation

    """

    opts = configuration.clean_opts(opts)

    utilities.check_dir(opts.get("output_dir"))

    log_file = os.path.join(opts.get("output_dir"), "{0}.log".format(opts.get("prefix")))
    utilities.configure_logger(filename=log_file, stream_level=opts.get("logger_stream_level"), file_level=opts.get("logger_file_level"))
    logger.debug("Logger configured")

    all_pockets, output_opts = pocket(**opts)

    return all_pockets, output_opts


def subpockets(bounding_spheres, ref_spheres, **opts):
    """

    Args:
      bounding_spheres (Spheres): a Spheres object containing both the peptide and solvent exposed face external spheres
      ref_spheres (Spheres): a Spheres object holding the interior spheres that define the pocket to be subdivided
      opts (dict): a dictionary containing all PyVOL options (see pyvol.configuration.clean_opts for details)

    Returns:
      grouped_list ([Spheres]): a list of Spheres objects each of which contains the geometric information describing a distinct subpocket

    """

    nonextraneous_rad = opts.get("min_rad") + opts.get("max_rad") + opts.get("inclusion_radius_buffer")
    nonextraneous_spheres = bounding_spheres.identify_nonextraneous(ref_spheres=ref_spheres, radius=nonextraneous_rad)

    sampling_radii = np.flip(np.arange(opts.get("min_rad"), opts.get("max_subpocket_rad"), opts.get("radial_sampling")), axis=0)
    unmerged_sphere_lists = utilities.sphere_multiprocessing(nonextraneous_spheres, sampling_radii, all_components=True)
    spheres = cluster.merge_sphere_list(itertools.chain(*unmerged_sphere_lists))

    cluster.hierarchically_cluster_spheres(spheres, ordered_radii=sampling_radii, min_new_radius=opts.get("min_subpocket_rad"), min_cluster_size=opts.get("min_cluster_size"), max_clusters=opts.get("max_clusters"))

    cluster.remove_overlap(spheres, radii=sampling_radii, spacing=opts.get("radial_sampling"))
    cluster.remove_overlap(spheres)
    cluster.remove_interior(spheres)
    grouped_list = cluster.extract_groups(spheres, surf_radius=opts.get("min_subpocket_surf_rad"), prefix=opts.get("prefix"))
    return grouped_list


def write_cfg(**opts):
    """ write the processed configuration to file

    Args:
      output_dir (str): output directory, relative or absolute
      prefix (str): identifying prefix for the output files

    """

    utilities.check_dir(opts.get("output_dir"))
    configuration.opts_to_file(opts)


def write_report(all_pockets, **opts):
    """ Write a brief report of calculated volumes to file

    Args:
      all_pockets ([Spheres]): a list of Spheres objects each of which contains the complete information about a distinct pocket or subpocket
      output_dir (str): output directory, relative or absolute
      prefix (str): identifying prefix for output files

    """
    import os
    import pandas as pd

    utilities.check_dir(opts.get("output_dir"))

    rept_list = []

    for pocket in all_pockets:
        spheres_name = os.path.join(opts.get("output_dir"), "{0}.xyzrg".format(pocket.name))
        pocket.write(spheres_name)
        rept_list.append({"name": pocket.name,
                          "volume": pocket.mesh.volume
                          })
    rept_df = pd.DataFrame(rept_list)
    rept_name = os.path.join(opts.get("output_dir"), "{0}.rept".format(opts.get("prefix")))
    rept_df.to_csv(rept_name, index=False)
    logger.info("Report written to: {0}".format(rept_name))




""" Front facing PyMOL functions """

from . import configuration
from . import identify
from . import pymol_utilities
from .spheres import Spheres
from . import utilities
import logging
import numpy as np
import os
import shutil


logger = logging.getLogger(__name__)

try:
    from pymol import cgo, cmd, CmdException
except:
    logger.error("PyMOL not imported")


def display_pockets(pockets, **opts):
    """ Display a list of pockets

    Args:
      pockets ([Spheres]): list of spheres object to display
      opts (dict): a dictionary containing all PyVOL options (see pyvol.pymol_interface.pymol_pocket_cmdline for details)

    """

    opts["palette"] = pymol_utilities.construct_palette(color_list=opts.get("palette"), max_value=len(pockets))

    if len(pockets) == 0:
        logger.info("No pockets found to display.")

    for index, pocket in enumerate(pockets):
        logger.info("Pocket {0} ({1}) Volume: \t{2} A^3\t({3})".format(index, pocket.name, np.round(pocket.mesh.volume), opts.get("palette")[index]))
        pymol_utilities.display_spheres_object(pocket, pocket.name, state=1, color=opts.get("palette")[index], alpha=opts.get("alpha"), mode=opts.get("display_mode"))


def load_calculation_cmdline(data_dir, prefix=None, display_mode=None, palette=None, alpha=None):
    """ Loads a pocket from memory and displays it in PyMOL

    Args:
      data_dir (str): directory containing PyVOL output (by default ends in .pyvol)
      prefix (str): internal display name (Default value = None)
      display_mode (str): display mode (Default value = "solid")
      palette (str): comma-separated list of PyMOL color strings (Default value = None)
      alpha (float): transparency value (Default value = 1.0)

    """

    if not os.path.isdir(data_dir):
        if os.path.isfile(data_dir):
            data_dir = os.path.dirname(data_dir)
        else:
            logger.error("Ambiguous/unparseable data_dir input: {0}".format(data_dir))
            raise ValueError

    input_opts = {}
    if prefix is not None:
        input_opts["display_prefix"] = prefix
    if display_mode is not None:
        input_opts["display_mode"] = display_mode
    if palette is not None:
        input_opts["palette"] = palette
    if alpha is not None:
        input_opts["alpha"] = alpha


    pockets, opts = identify.load_calculation(data_dir, input_opts=input_opts)
    display_pockets(pockets, **opts)
    logger.info("Loading {0} with mode {1}".format(spheres.name, display_mode))


def pymol_pocket_cmdline(protein=None, ligand=None, prot_file=None, lig_file=None, min_rad=1.4, max_rad=3.4, constrain_radii=True, mode="largest", coordinates=None, residue=None, resid=None, lig_excl_rad=None, lig_incl_rad=None, min_volume=200, subdivide=False, max_clusters=None, min_subpocket_rad=1.7, max_subpocket_rad=3.4, min_subpocket_surf_rad=1.0, radial_sampling=0.1, inclusion_radius_buffer=1.0, min_cluster_size=50, project_dir=None, output_dir=None, prefix=None, logger_stream_level="INFO", logger_file_level="DEBUG", protein_only=False, display_mode="solid", alpha=1.0, palette=None):
    """ PyMOL-compatible command line entry point

    Args:
      protein (str): PyMOL-only PyMOL selection string for the protein (Default value = None)
      ligand (str): PyMOL-only PyMOL selection string for the ligand (Default value = None)
      prot_file (str): filename for the input pdb file containing the peptide--redundant with protein argument (Default value =- )
      lig_file (str): filename for the input pdb file containing a ligand--redundant with ligand argument (Default value = None)
      min_rad (float): radius for SES calculations (Default value = 1.4)
      max_rad (float): radius used to identify the outer, bulk solvent exposed surface (Default value = 3.4)
      constrain_radii (bool): restrict input radii to tested values? (Default value = False)
      mode (str): pocket identification mode (can be largest, all, or specific) (Default value = "largest")
      coordinates ([float]): 3D coordinate used for pocket specification (Default value = None)
      residue (str): Pymol-only PyMOL selection string for a residue to use for pocket specification (Default value=None)
      resid (str): residue identifier for pocket specification (Default value = None)
      lig_excl_rad (float): maximum distance from a provided ligand that can be included in calculated pockets (Default value = None)
      lig_incl_rad (float): minimum distance from a provided ligand that should be included in calculated pockets when solvent border is ambiguous (Default value = None)
      min_volume (float): minimum volume of pockets returned when running in 'all' mode (Default value = 200)
      subdivide (bool): calculate subpockets? (Default value = False)
      max_clusters (int): maximum number of clusters (Default value = None)
      min_subpocket_rad (float): minimum radius that identifies distinct subpockets (Default value = 1.7)
      max_subpocket_rad (float): maximum sampling radius used in subpocket identification (Default value = 3.4)
      min_subpocket_surf_rad (float): radius used to calculate subpocket surfaces (Default value = 1.0)
      inclusion_radius_buffer (float): buffer radius in excess of the nonextraneous radius from the identified pocket used to identify atoms pertinent to subpocket clustering (Default value = 1.0)
      radial_sampling (float): radial sampling used for subpocket clustering (Default value = 0.1)
      min_cluster_size (int): minimum number of spheres in a proper cluster; used to eliminate insignificant subpockets (Default value = 50)
      project_dir (str): parent directory in which to create the output directory if the output directory is unspecified (Default value = None)
      output_dir (str): filename of the directory in which to place all output; can be absolute or relative (Default value = None)
      prefix (str): identifying string for output (Default value = None)
      logger_stream_level (str): sets the logger level for stdio output (Default value = "INFO")
      logger_file_level (str): sets the logger level for file output (Default value = "DEBUG")
      protein_only (bool): PyMOL-only include only peptides in protein file
      display_mode (str): PyMOL-only display mode for calculated pockets (Default value = "solid")
      alpha (float): PyMOL-only display option specifying translucency of CGO objects (Default value = 1.0)
      palette (str): PyMOL-only display option representing a comma separated list of PyMOL color strings (Default value = None)

    """

    opts = {
        "protein": protein,
        "ligand": ligand,
        "prot_file": prot_file,
        "lig_file": lig_file,
        "min_rad": min_rad,
        "max_rad": max_rad,
        "constrain_radii": constrain_radii,
        "mode": mode,
        "residue": residue,
        "resid": resid,
        "coordinates": coordinates,
        "lig_excl_rad": lig_excl_rad,
        "lig_incl_rad": lig_incl_rad,
        "min_volume": min_volume,
        "subdivide": subdivide,
        "max_clusters": max_clusters,
        "min_subpocket_rad": min_subpocket_rad,
        "max_subpocket_rad": max_subpocket_rad,
        "min_subpocket_surf_rad": min_subpocket_surf_rad,
        "radial_sampling": radial_sampling,
        "inclusion_radius_buffer": inclusion_radius_buffer,
        "min_cluster_size": min_cluster_size,
        "project_dir": project_dir,
        "output_dir": output_dir,
        "prefix": prefix,
        "logger_stream_level": logger_stream_level,
        "logger_file_level": logger_file_level,
        "protein_only": protein_only,
        "display_mode": display_mode,
        "alpha": alpha,
        "palette": palette
    }

    pymol_pocket(**opts)

def pymol_pocket(**opts):
    """ Perform PyMOL-dependent processing of inputs to generate input files for PyVOL pocket processing

    Args:
      opts (dict): dictionary containing all PyVOL options (see pyvol.pymol_interface.pymol_pocket_cmdline for details)

    Returns:
      pockets ([Spheres]): a list of Spheres objects each of which contains the geometric information describing a distinct pocket or subpocket
      output_opts (dict): dictionary containing the actual options used in the pocket calculation

    """


    boolean_args = ["constrain_radii", "subdivide", "protein_only"]
    for arg in boolean_args:
        if not isinstance(opts.get(arg), bool):
            if opts.get(arg) in ["True", "true", "t", "1"]:
                opts[arg] = True
            elif opts.get(arg) in ["False", "false", "f", "0"]:
                opts[arg] = False
            else:
                logger.warning("Boolean argument {0} ({1}) not parsed correctly and reverting to default".format(arg, opts[arg]))

    opts = configuration.clean_opts(opts)

    if opts.get("protein") is None:
        if opts.get("prot_file") is None:
            logger.error("No protein input: prot_file and protein inputs are empty")
            raise ValueError
        else:
            logger.debug("Protein file already specified on disk; skipping protein processing.")
    else:
        if opts.get("protein_only"):
            opts["protein"] = "({0}) and (poly)".format(opts.get("protein"))

        if opts.get("ligand") is not None:
            opts["protein"] = "({0}) and not ({1})".format(opts.get("protein"), opts.get("ligand"))

        logger.debug("Final protein selection: {0}".format(opts.get("protein")))
        prot_atoms = cmd.count_atoms(opts.get("protein"))
        if prot_atoms == 0:
            logger.error("No atoms included in protein selection--ending calculation")
            return
        elif prot_atoms < 100:
            logger.warning("Only {0} atoms included in protein selection".format(prot_atoms))

        cmd.save(opts.get("prot_file"), opts.get("protein"))
        logger.debug("Protein '{0}' saved to {1}".format(opts.get("protein"), opts.get("prot_file")))

    if opts.get("ligand") is not None:
        cmd.save(opts.get("lig_file"), opts.get("ligand"))
        logger.debug("Ligand selection: {0}".format(opts.get("ligand")))

    if opts.get("coordinates") is not None:
        opts["residue"] = None
    else:
        if opts.get("residue") is not None:
            opts["coordinates"] = cmd.get_coords("{0} and sidechain".format(opts.get("residue")), 1)

    pockets, output_opts = identify.pocket_wrapper(**opts)

    display_pockets(pockets, **output_opts)
    return pockets, output_opts




import argparse
from pyvol import configuration, identify


if __name__ == "__main__":
    parser = argparse.ArgumentParser(prog="PyVOL", description="Identification, calculation, and segmentation of protein binding pockets", epilog="For complete documentation and tutorial on use, visit the project webpage: https://schlessingerlab.github.io/pyvol")
    parser.add_argument("cfg_file", help="input configuration file or output configuration file if specificying --template option")
    parser.add_argument("-t", "--template", action='store_true', help="write a template configuration file")

    args = parser.parse_args()

    if args.template:
        configuration.defaults_to_file(args.cfg_file)
    else:
        identify.pocket_wrapper(**configuration.file_to_opts(args.cfg_file))




import pathlib
from setuptools import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name="bio-pyvol",
    version="1.6.7",
    description="a PyMOL plugin and python package for visualization, comparison, and volume calculation of protein drug-binding sites",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/schlessingerlab/pyvol",
    author="Ryan H.B. Smith",
    author_email="ryan.smith@icahn.mssm.edu",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        ],
    packages=["pyvol"],
    install_requires=[
        "biopython>=1.73",
        "numpy>=1.16.1",
        "pandas>=0.24.1",
        "scipy>=1.2.1",
        "scikit-learn>=0.20.2",
        "trimesh>=2.36.29",
        "configparser",
        ],
    entry_points={
        "console_scripts": [
            "pyvol=pyvol.__main__:main",
            ]
        },
    include_package_data=True,
    package_data={
        'pyvol': ['pkgs/msms_2.6.1/msms.*']
        },
    )
